@@ -1,38 +1,40 @@
+// Copyright (c) The Thanos Authors.
+// Licensed under the Apache License 2.0.
+
 package store
 
 import (
 	"context"
+	"fmt"
 	"io/ioutil"
 	"os"
 	"path/filepath"
-	"sync"
+	"strings"
 	"testing"
 	"time"
 
 	"github.com/go-kit/kit/log"
 	"github.com/oklog/ulid"
-	"github.com/pkg/errors"
+	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/relabel"
 	"github.com/prometheus/prometheus/pkg/timestamp"
-	"github.com/prometheus/prometheus/tsdb/labels"
 	"github.com/thanos-io/thanos/pkg/block"
 	"github.com/thanos-io/thanos/pkg/block/metadata"
 	"github.com/thanos-io/thanos/pkg/model"
 	"github.com/thanos-io/thanos/pkg/objstore"
-	"github.com/thanos-io/thanos/pkg/objstore/inmem"
 	"github.com/thanos-io/thanos/pkg/objstore/objtesting"
-	"github.com/thanos-io/thanos/pkg/runutil"
 	storecache "github.com/thanos-io/thanos/pkg/store/cache"
 	"github.com/thanos-io/thanos/pkg/store/storepb"
 	"github.com/thanos-io/thanos/pkg/testutil"
+	"github.com/thanos-io/thanos/pkg/testutil/e2eutil"
 )
 
 var (
-	minTime         = time.Unix(0, 0)
-	maxTime, _      = time.Parse(time.RFC3339, "9999-12-31T23:59:59Z")
-	minTimeDuration = model.TimeOrDurationValue{Time: &minTime}
-	maxTimeDuration = model.TimeOrDurationValue{Time: &maxTime}
-	filterConf      = &FilterConfig{
+	minTime            = time.Unix(0, 0)
+	maxTime, _         = time.Parse(time.RFC3339, "9999-12-31T23:59:59Z")
+	minTimeDuration    = model.TimeOrDurationValue{Time: &minTime}
+	maxTimeDuration    = model.TimeOrDurationValue{Time: &maxTime}
+	allowAllFilterConf = &FilterConfig{
 		MinTime: minTimeDuration,
 		MaxTime: maxTimeDuration,
 	}
@@ -40,39 +42,41 @@ var (
 
 type noopCache struct{}
 
-func (noopCache) SetPostings(b ulid.ULID, l labels.Label, v []byte)   {}
-func (noopCache) Postings(b ulid.ULID, l labels.Label) ([]byte, bool) { return nil, false }
-func (noopCache) SetSeries(b ulid.ULID, id uint64, v []byte)          {}
-func (noopCache) Series(b ulid.ULID, id uint64) ([]byte, bool)        { return nil, false }
-
-type swappableCache struct {
-	ptr indexCache
+func (noopCache) StorePostings(ctx context.Context, blockID ulid.ULID, l labels.Label, v []byte) {}
+func (noopCache) FetchMultiPostings(ctx context.Context, blockID ulid.ULID, keys []labels.Label) (map[labels.Label][]byte, []labels.Label) {
+	return map[labels.Label][]byte{}, keys
 }
 
-func (c *swappableCache) SwapWith(ptr2 indexCache) {
+func (noopCache) StoreSeries(ctx context.Context, blockID ulid.ULID, id uint64, v []byte) {}
+func (noopCache) FetchMultiSeries(ctx context.Context, blockID ulid.ULID, ids []uint64) (map[uint64][]byte, []uint64) {
+	return map[uint64][]byte{}, ids
+}
+
+type swappableCache struct {
+	ptr storecache.IndexCache
+}
+
+func (c *swappableCache) SwapWith(ptr2 storecache.IndexCache) {
 	c.ptr = ptr2
 }
 
-func (c *swappableCache) SetPostings(b ulid.ULID, l labels.Label, v []byte) {
-	c.ptr.SetPostings(b, l, v)
+func (c *swappableCache) StorePostings(ctx context.Context, blockID ulid.ULID, l labels.Label, v []byte) {
+	c.ptr.StorePostings(ctx, blockID, l, v)
 }
 
-func (c *swappableCache) Postings(b ulid.ULID, l labels.Label) ([]byte, bool) {
-	return c.ptr.Postings(b, l)
+func (c *swappableCache) FetchMultiPostings(ctx context.Context, blockID ulid.ULID, keys []labels.Label) (map[labels.Label][]byte, []labels.Label) {
+	return c.ptr.FetchMultiPostings(ctx, blockID, keys)
 }
 
-func (c *swappableCache) SetSeries(b ulid.ULID, id uint64, v []byte) {
-	c.ptr.SetSeries(b, id, v)
+func (c *swappableCache) StoreSeries(ctx context.Context, blockID ulid.ULID, id uint64, v []byte) {
+	c.ptr.StoreSeries(ctx, blockID, id, v)
 }
 
-func (c *swappableCache) Series(b ulid.ULID, id uint64) ([]byte, bool) {
-	return c.ptr.Series(b, id)
+func (c *swappableCache) FetchMultiSeries(ctx context.Context, blockID ulid.ULID, ids []uint64) (map[uint64][]byte, []uint64) {
+	return c.ptr.FetchMultiSeries(ctx, blockID, ids)
 }
 
 type storeSuite struct {
-	cancel context.CancelFunc
-	wg     sync.WaitGroup
-
 	store            *BucketStore
 	minTime, maxTime int64
 	cache            *swappableCache
@@ -80,13 +84,8 @@ type storeSuite struct {
 	logger log.Logger
 }
 
-func (s *storeSuite) Close() {
-	s.cancel()
-	s.wg.Wait()
-}
-
 func prepareTestBlocks(t testing.TB, now time.Time, count int, dir string, bkt objstore.Bucket,
-	series []labels.Labels, extLset labels.Labels) (blocks int, minTime, maxTime int64) {
+	series []labels.Labels, extLset labels.Labels) (minTime, maxTime int64) {
 	ctx := context.Background()
 	logger := log.NewNopLogger()
 
@@ -102,9 +101,9 @@ func prepareTestBlocks(t testing.TB, now time.Time, count int, dir string, bkt o
 
 		// Create two blocks per time slot. Only add 10 samples each so only one chunk
 		// gets created each. This way we can easily verify we got 10 chunks per series below.
-		id1, err := testutil.CreateBlock(ctx, dir, series[:4], 10, mint, maxt, extLset, 0)
+		id1, err := e2eutil.CreateBlock(ctx, dir, series[:4], 10, mint, maxt, extLset, 0)
 		testutil.Ok(t, err)
-		id2, err := testutil.CreateBlock(ctx, dir, series[4:], 10, mint, maxt, extLset, 0)
+		id2, err := e2eutil.CreateBlock(ctx, dir, series[4:], 10, mint, maxt, extLset, 0)
 		testutil.Ok(t, err)
 
 		dir1, dir2 := filepath.Join(dir, id1.String()), filepath.Join(dir, id2.String())
@@ -117,7 +116,6 @@ func prepareTestBlocks(t testing.TB, now time.Time, count int, dir string, bkt o
 
 		testutil.Ok(t, block.Upload(ctx, logger, bkt, dir1))
 		testutil.Ok(t, block.Upload(ctx, logger, bkt, dir2))
-		blocks += 2
 
 		testutil.Ok(t, os.RemoveAll(dir1))
 		testutil.Ok(t, os.RemoveAll(dir2))
@@ -126,7 +124,7 @@ func prepareTestBlocks(t testing.TB, now time.Time, count int, dir string, bkt o
 	return
 }
 
-func prepareStoreWithTestBlocks(t testing.TB, dir string, bkt objstore.Bucket, manyParts bool, maxSampleCount uint64, relabelConfig []*relabel.Config) *storeSuite {
+func prepareStoreWithTestBlocks(t testing.TB, dir string, bkt objstore.Bucket, manyParts bool, maxChunksLimit uint64, relabelConfig []*relabel.Config, filterConf *FilterConfig) *storeSuite {
 	series := []labels.Labels{
 		labels.FromStrings("a", "1", "b", "1"),
 		labels.FromStrings("a", "1", "b", "2"),
@@ -139,19 +137,40 @@ func prepareStoreWithTestBlocks(t testing.TB, dir string, bkt objstore.Bucket, m
 	}
 	extLset := labels.FromStrings("ext1", "value1")
 
-	blocks, minTime, maxTime := prepareTestBlocks(t, time.Now(), 3, dir, bkt,
+	minTime, maxTime := prepareTestBlocks(t, time.Now(), 3, dir, bkt,
 		series, extLset)
 
-	ctx, cancel := context.WithCancel(context.Background())
 	s := &storeSuite{
-		cancel:  cancel,
 		logger:  log.NewLogfmtLogger(os.Stderr),
 		cache:   &swappableCache{},
 		minTime: minTime,
 		maxTime: maxTime,
 	}
 
-	store, err := NewBucketStore(s.logger, nil, bkt, dir, s.cache, 0, maxSampleCount, 20, false, 20, filterConf, relabelConfig, true)
+	metaFetcher, err := block.NewMetaFetcher(s.logger, 20, objstore.WithNoopInstr(bkt), dir, nil, []block.MetadataFilter{
+		block.NewTimePartitionMetaFilter(filterConf.MinTime, filterConf.MaxTime),
+		block.NewLabelShardedMetaFilter(relabelConfig),
+	}, nil)
+	testutil.Ok(t, err)
+
+	store, err := NewBucketStore(
+		s.logger,
+		nil,
+		objstore.WithNoopInstr(bkt),
+		metaFetcher,
+		dir,
+		s.cache,
+		nil,
+		0,
+		NewChunksLimiterFactory(maxChunksLimit),
+		false,
+		20,
+		filterConf,
+		true,
+		true,
+		DefaultPostingOffsetInMemorySampling,
+		true,
+	)
 	testutil.Ok(t, err)
 	s.store = store
 
@@ -159,43 +178,32 @@ func prepareStoreWithTestBlocks(t testing.TB, dir string, bkt objstore.Bucket, m
 		s.store.partitioner = naivePartitioner{}
 	}
 
-	s.wg.Add(1)
-	go func() {
-		defer s.wg.Done()
-
-		if err := runutil.Repeat(100*time.Millisecond, ctx.Done(), func() error {
-			return store.SyncBlocks(ctx)
-		}); err != nil && errors.Cause(err) != context.Canceled {
-			t.Fatal(err)
-		}
-	}()
-
-	rctx, rcancel := context.WithTimeout(ctx, 30*time.Second)
-	defer rcancel()
-
-	testutil.Ok(t, runutil.Retry(100*time.Millisecond, rctx.Done(), func() error {
-		if store.numBlocks() < blocks {
-			return errors.New("not all blocks loaded")
-		}
-		return nil
-	}))
+	ctx, cancel := context.WithTimeout(context.Background(), 30*time.Second)
+	defer cancel()
 
+	testutil.Ok(t, store.SyncBlocks(ctx))
 	return s
 }
 
-func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
+// TODO(bwplotka): Benchmark Series.
+func testBucketStore_e2e(t *testing.T, ctx context.Context, s *storeSuite) {
 	mint, maxt := s.store.TimeRange()
 	testutil.Equals(t, s.minTime, mint)
 	testutil.Equals(t, s.maxTime, maxt)
 
-	vals, err := s.store.LabelValues(ctx, &storepb.LabelValuesRequest{Label: "a"})
+	vals, err := s.store.LabelValues(ctx, &storepb.LabelValuesRequest{
+		Label: "a",
+		Start: timestamp.FromTime(minTime),
+		End:   timestamp.FromTime(maxTime),
+	})
 	testutil.Ok(t, err)
 	testutil.Equals(t, []string{"1", "2"}, vals.Values)
 
 	// TODO(bwplotka): Add those test cases to TSDB querier_test.go as well, there are no tests for matching.
 	for i, tcase := range []struct {
-		req      *storepb.SeriesRequest
-		expected [][]storepb.Label
+		req              *storepb.SeriesRequest
+		expected         [][]storepb.Label
+		expectedChunkLen int
 	}{
 		{
 			req: &storepb.SeriesRequest{
@@ -205,6 +213,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -224,6 +233,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -239,6 +249,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -254,6 +265,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -273,6 +285,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -292,6 +305,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "2"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -307,6 +321,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "c", Value: "1"}, {Name: "ext2", Value: "value2"}},
 				{{Name: "a", Value: "1"}, {Name: "c", Value: "2"}, {Name: "ext2", Value: "value2"}},
@@ -330,6 +345,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -345,6 +361,7 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MinTime: mint,
 				MaxTime: maxt,
 			},
+			expectedChunkLen: 3,
 			expected: [][]storepb.Label{
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
 				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
@@ -368,23 +385,43 @@ func testBucketStore_e2e(t testing.TB, ctx context.Context, s *storeSuite) {
 				MaxTime: maxt,
 			},
 		},
+		// Test skip-chunk option.
+		{
+			req: &storepb.SeriesRequest{
+				Matchers: []storepb.LabelMatcher{
+					{Type: storepb.LabelMatcher_EQ, Name: "a", Value: "1"},
+				},
+				MinTime:    mint,
+				MaxTime:    maxt,
+				SkipChunks: true,
+			},
+			expectedChunkLen: 0,
+			expected: [][]storepb.Label{
+				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
+				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
+				{{Name: "a", Value: "1"}, {Name: "c", Value: "1"}, {Name: "ext2", Value: "value2"}},
+				{{Name: "a", Value: "1"}, {Name: "c", Value: "2"}, {Name: "ext2", Value: "value2"}},
+			},
+		},
 	} {
-		t.Log("Run ", i)
+		if ok := t.Run(fmt.Sprint(i), func(t *testing.T) {
+			srv := newStoreSeriesServer(ctx)
 
-		srv := newStoreSeriesServer(ctx)
+			testutil.Ok(t, s.store.Series(tcase.req, srv))
+			testutil.Equals(t, len(tcase.expected), len(srv.SeriesSet))
 
-		testutil.Ok(t, s.store.Series(tcase.req, srv))
-		testutil.Equals(t, len(tcase.expected), len(srv.SeriesSet))
-
-		for i, s := range srv.SeriesSet {
-			testutil.Equals(t, tcase.expected[i], s.Labels)
-			testutil.Equals(t, 3, len(s.Chunks))
+			for i, s := range srv.SeriesSet {
+				testutil.Equals(t, tcase.expected[i], s.Labels)
+				testutil.Equals(t, tcase.expectedChunkLen, len(s.Chunks))
+			}
+		}); !ok {
+			return
 		}
 	}
 }
 
 func TestBucketStore_e2e(t *testing.T) {
-	objtesting.ForeachStore(t, func(t testing.TB, bkt objstore.Bucket) {
+	objtesting.ForeachStore(t, func(t *testing.T, bkt objstore.Bucket) {
 		ctx, cancel := context.WithCancel(context.Background())
 		defer cancel()
 
@@ -392,30 +429,36 @@ func TestBucketStore_e2e(t *testing.T) {
 		testutil.Ok(t, err)
 		defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
 
-		s := prepareStoreWithTestBlocks(t, dir, bkt, false, 0, emptyRelabelConfig)
-		defer s.Close()
+		s := prepareStoreWithTestBlocks(t, dir, bkt, false, 0, emptyRelabelConfig, allowAllFilterConf)
 
-		t.Log("Test with no index cache")
-		s.cache.SwapWith(noopCache{})
-		testBucketStore_e2e(t, ctx, s)
+		if ok := t.Run("no index cache", func(t *testing.T) {
+			s.cache.SwapWith(noopCache{})
+			testBucketStore_e2e(t, ctx, s)
+		}); !ok {
+			return
+		}
 
-		t.Log("Test with large, sufficient index cache")
-		indexCache, err := storecache.NewIndexCache(s.logger, nil, storecache.Opts{
-			MaxItemSizeBytes: 1e5,
-			MaxSizeBytes:     2e5,
+		if ok := t.Run("with large, sufficient index cache", func(t *testing.T) {
+			indexCache, err := storecache.NewInMemoryIndexCacheWithConfig(s.logger, nil, storecache.InMemoryIndexCacheConfig{
+				MaxItemSize: 1e5,
+				MaxSize:     2e5,
+			})
+			testutil.Ok(t, err)
+			s.cache.SwapWith(indexCache)
+			testBucketStore_e2e(t, ctx, s)
+		}); !ok {
+			return
+		}
+
+		t.Run("with small index cache", func(t *testing.T) {
+			indexCache2, err := storecache.NewInMemoryIndexCacheWithConfig(s.logger, nil, storecache.InMemoryIndexCacheConfig{
+				MaxItemSize: 50,
+				MaxSize:     100,
+			})
+			testutil.Ok(t, err)
+			s.cache.SwapWith(indexCache2)
+			testBucketStore_e2e(t, ctx, s)
 		})
-		testutil.Ok(t, err)
-		s.cache.SwapWith(indexCache)
-		testBucketStore_e2e(t, ctx, s)
-
-		t.Log("Test with small index cache")
-		indexCache2, err := storecache.NewIndexCache(s.logger, nil, storecache.Opts{
-			MaxItemSizeBytes: 50,
-			MaxSizeBytes:     100,
-		})
-		testutil.Ok(t, err)
-		s.cache.SwapWith(indexCache2)
-		testBucketStore_e2e(t, ctx, s)
 	})
 }
 
@@ -433,7 +476,7 @@ func (g naivePartitioner) Partition(length int, rng func(int) (uint64, uint64))
 // This tests if our, sometimes concurrent, fetches for different parts works.
 // Regression test against: https://github.com/thanos-io/thanos/issues/829.
 func TestBucketStore_ManyParts_e2e(t *testing.T) {
-	objtesting.ForeachStore(t, func(t testing.TB, bkt objstore.Bucket) {
+	objtesting.ForeachStore(t, func(t *testing.T, bkt objstore.Bucket) {
 		ctx, cancel := context.WithCancel(context.Background())
 		defer cancel()
 
@@ -441,12 +484,11 @@ func TestBucketStore_ManyParts_e2e(t *testing.T) {
 		testutil.Ok(t, err)
 		defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
 
-		s := prepareStoreWithTestBlocks(t, dir, bkt, true, 0, emptyRelabelConfig)
-		defer s.Close()
+		s := prepareStoreWithTestBlocks(t, dir, bkt, true, 0, emptyRelabelConfig, allowAllFilterConf)
 
-		indexCache, err := storecache.NewIndexCache(s.logger, nil, storecache.Opts{
-			MaxItemSizeBytes: 1e5,
-			MaxSizeBytes:     2e5,
+		indexCache, err := storecache.NewInMemoryIndexCacheWithConfig(s.logger, nil, storecache.InMemoryIndexCacheConfig{
+			MaxItemSize: 1e5,
+			MaxSize:     2e5,
 		})
 		testutil.Ok(t, err)
 		s.cache.SwapWith(indexCache)
@@ -458,75 +500,172 @@ func TestBucketStore_ManyParts_e2e(t *testing.T) {
 func TestBucketStore_TimePartitioning_e2e(t *testing.T) {
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
-	bkt := inmem.NewBucket()
+	bkt := objstore.NewInMemBucket()
 
 	dir, err := ioutil.TempDir("", "test_bucket_time_part_e2e")
 	testutil.Ok(t, err)
 	defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
 
-	series := []labels.Labels{
-		labels.FromStrings("a", "1", "b", "1"),
-		labels.FromStrings("a", "1", "b", "1"),
-		labels.FromStrings("a", "1", "b", "1"),
-		labels.FromStrings("a", "1", "b", "1"),
-		labels.FromStrings("a", "1", "b", "2"),
-		labels.FromStrings("a", "1", "b", "2"),
-		labels.FromStrings("a", "1", "b", "2"),
-		labels.FromStrings("a", "1", "b", "2"),
-	}
-	extLset := labels.FromStrings("ext1", "value1")
-
-	_, minTime, _ := prepareTestBlocks(t, time.Now(), 3, dir, bkt, series, extLset)
-
 	hourAfter := time.Now().Add(1 * time.Hour)
 	filterMaxTime := model.TimeOrDurationValue{Time: &hourAfter}
 
-	store, err := NewBucketStore(nil, nil, bkt, dir, noopCache{}, 0, 0, 20, false, 20,
-		&FilterConfig{
-			MinTime: minTimeDuration,
-			MaxTime: filterMaxTime,
-		}, emptyRelabelConfig, true)
-	testutil.Ok(t, err)
+	// The query will fetch 2 series from 2 blocks, so we do expect to hit a total of 4 chunks.
+	expectedChunks := uint64(2 * 2)
 
-	err = store.SyncBlocks(ctx)
-	testutil.Ok(t, err)
+	s := prepareStoreWithTestBlocks(t, dir, bkt, false, expectedChunks, emptyRelabelConfig, &FilterConfig{
+		MinTime: minTimeDuration,
+		MaxTime: filterMaxTime,
+	})
+	testutil.Ok(t, s.store.SyncBlocks(ctx))
 
-	mint, maxt := store.TimeRange()
-	testutil.Equals(t, minTime, mint)
+	mint, maxt := s.store.TimeRange()
+	testutil.Equals(t, s.minTime, mint)
 	testutil.Equals(t, filterMaxTime.PrometheusTimestamp(), maxt)
 
-	for i, tcase := range []struct {
-		req            *storepb.SeriesRequest
-		expectedLabels [][]storepb.Label
-		expectedChunks int
+	req := &storepb.SeriesRequest{
+		Matchers: []storepb.LabelMatcher{
+			{Type: storepb.LabelMatcher_EQ, Name: "a", Value: "1"},
+		},
+		MinTime: mint,
+		MaxTime: timestamp.FromTime(time.Now().AddDate(0, 0, 1)),
+	}
+
+	expectedLabels := [][]storepb.Label{
+		{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
+		{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext1", Value: "value1"}},
+		{{Name: "a", Value: "1"}, {Name: "c", Value: "1"}, {Name: "ext2", Value: "value2"}},
+		{{Name: "a", Value: "1"}, {Name: "c", Value: "2"}, {Name: "ext2", Value: "value2"}},
+	}
+
+	s.cache.SwapWith(noopCache{})
+	srv := newStoreSeriesServer(ctx)
+
+	testutil.Ok(t, s.store.Series(req, srv))
+	testutil.Equals(t, len(expectedLabels), len(srv.SeriesSet))
+
+	for i, s := range srv.SeriesSet {
+		testutil.Equals(t, expectedLabels[i], s.Labels)
+
+		// prepareTestBlocks makes 3 chunks containing 2 hour data,
+		// we should only get 1, as we are filtering by time.
+		testutil.Equals(t, 1, len(s.Chunks))
+	}
+}
+
+func TestBucketStore_Series_ChunksLimiter_e2e(t *testing.T) {
+	// The query will fetch 2 series from 6 blocks, so we do expect to hit a total of 12 chunks.
+	expectedChunks := uint64(2 * 6)
+
+	cases := map[string]struct {
+		maxChunksLimit uint64
+		expectedErr    string
 	}{
-		{
-			req: &storepb.SeriesRequest{
+		"should succeed if the max chunks limit is not exceeded": {
+			maxChunksLimit: expectedChunks,
+		},
+		"should fail if the max chunks limit is exceeded": {
+			maxChunksLimit: expectedChunks - 1,
+			expectedErr:    "exceeded chunks limit",
+		},
+	}
+
+	for testName, testData := range cases {
+		t.Run(testName, func(t *testing.T) {
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+			bkt := objstore.NewInMemBucket()
+
+			dir, err := ioutil.TempDir("", "test_bucket_chunks_limiter_e2e")
+			testutil.Ok(t, err)
+			defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
+
+			s := prepareStoreWithTestBlocks(t, dir, bkt, false, testData.maxChunksLimit, emptyRelabelConfig, allowAllFilterConf)
+			testutil.Ok(t, s.store.SyncBlocks(ctx))
+
+			req := &storepb.SeriesRequest{
 				Matchers: []storepb.LabelMatcher{
 					{Type: storepb.LabelMatcher_EQ, Name: "a", Value: "1"},
 				},
-				MinTime: mint,
-				MaxTime: timestamp.FromTime(time.Now().AddDate(0, 0, 1)),
-			},
-			expectedLabels: [][]storepb.Label{
-				{{Name: "a", Value: "1"}, {Name: "b", Value: "1"}, {Name: "ext1", Value: "value1"}},
-				{{Name: "a", Value: "1"}, {Name: "b", Value: "2"}, {Name: "ext2", Value: "value2"}},
-			},
-			// prepareTestBlocks makes 3 chunks containing 2 hour data,
-			// we should only get 1, as we are filtering.
-			expectedChunks: 1,
-		},
-	} {
-		t.Log("Run", i)
+				MinTime: minTimeDuration.PrometheusTimestamp(),
+				MaxTime: maxTimeDuration.PrometheusTimestamp(),
+			}
 
-		srv := newStoreSeriesServer(ctx)
+			s.cache.SwapWith(noopCache{})
+			srv := newStoreSeriesServer(ctx)
+			err = s.store.Series(req, srv)
 
-		testutil.Ok(t, store.Series(tcase.req, srv))
-		testutil.Equals(t, len(tcase.expectedLabels), len(srv.SeriesSet))
-
-		for i, s := range srv.SeriesSet {
-			testutil.Equals(t, tcase.expectedLabels[i], s.Labels)
-			testutil.Equals(t, tcase.expectedChunks, len(s.Chunks))
-		}
+			if testData.expectedErr == "" {
+				testutil.Ok(t, err)
+			} else {
+				testutil.NotOk(t, err)
+				testutil.Assert(t, strings.Contains(err.Error(), testData.expectedErr))
+			}
+		})
 	}
 }
+
+func TestBucketStore_LabelNames_e2e(t *testing.T) {
+	objtesting.ForeachStore(t, func(t *testing.T, bkt objstore.Bucket) {
+		ctx, cancel := context.WithCancel(context.Background())
+		defer cancel()
+
+		dir, err := ioutil.TempDir("", "test_bucketstore_label_names_e2e")
+		testutil.Ok(t, err)
+		defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
+
+		s := prepareStoreWithTestBlocks(t, dir, bkt, false, 0, emptyRelabelConfig, allowAllFilterConf)
+
+		mint, maxt := s.store.TimeRange()
+		testutil.Equals(t, s.minTime, mint)
+		testutil.Equals(t, s.maxTime, maxt)
+
+		vals, err := s.store.LabelNames(ctx, &storepb.LabelNamesRequest{
+			Start: timestamp.FromTime(minTime),
+			End:   timestamp.FromTime(maxTime),
+		})
+		testutil.Ok(t, err)
+		testutil.Equals(t, []string{"a", "b", "c"}, vals.Names)
+
+		// Outside the time range.
+		vals, err = s.store.LabelNames(ctx, &storepb.LabelNamesRequest{
+			Start: timestamp.FromTime(time.Now().Add(-24 * time.Hour)),
+			End:   timestamp.FromTime(time.Now().Add(-23 * time.Hour)),
+		})
+		testutil.Ok(t, err)
+		testutil.Equals(t, []string(nil), vals.Names)
+	})
+}
+
+func TestBucketStore_LabelValues_e2e(t *testing.T) {
+	objtesting.ForeachStore(t, func(t *testing.T, bkt objstore.Bucket) {
+		ctx, cancel := context.WithCancel(context.Background())
+		defer cancel()
+
+		dir, err := ioutil.TempDir("", "test_bucketstore_label_values_e2e")
+		testutil.Ok(t, err)
+		defer func() { testutil.Ok(t, os.RemoveAll(dir)) }()
+
+		s := prepareStoreWithTestBlocks(t, dir, bkt, false, 0, emptyRelabelConfig, allowAllFilterConf)
+
+		mint, maxt := s.store.TimeRange()
+		testutil.Equals(t, s.minTime, mint)
+		testutil.Equals(t, s.maxTime, maxt)
+
+		vals, err := s.store.LabelValues(ctx, &storepb.LabelValuesRequest{
+			Label: "a",
+			Start: timestamp.FromTime(minTime),
+			End:   timestamp.FromTime(maxTime),
+		})
+		testutil.Ok(t, err)
+		testutil.Equals(t, []string{"1", "2"}, vals.Values)
+
+		// Outside the time range.
+		vals, err = s.store.LabelValues(ctx, &storepb.LabelValuesRequest{
+			Label: "a",
+			Start: timestamp.FromTime(time.Now().Add(-24 * time.Hour)),
+			End:   timestamp.FromTime(time.Now().Add(-23 * time.Hour)),
+		})
+		testutil.Ok(t, err)
+		testutil.Equals(t, []string(nil), vals.Values)
+	})
+}
